{-# LANGUAGE OverloadedStrings #-}

module Windows12.Lexer where

import Data.Text (Text)
import qualified Data.Text as T
import Data.Void
import Text.Megaparsec
import Text.Megaparsec.Char
import qualified Text.Megaparsec.Char.Lexer as L

type Parser = Parsec Void Text

sc :: Parser ()
sc =
  L.space
    space1
    (L.skipLineComment "#")
    (L.skipBlockComment "/*" "*/")

lexeme :: Parser a -> Parser a
lexeme = L.lexeme sc

symbol :: Text -> Parser Text
symbol = L.symbol sc

charLiteral :: Parser Char
charLiteral = between (char '\'') (char '\'') L.charLiteral

stringLiteral :: Parser Text
stringLiteral = T.pack <$> (char '"' *> manyTill L.charLiteral (char '"'))

intLiteral :: Parser Int
intLiteral = lexeme L.decimal

uintLiteral :: Parser Word
uintLiteral = lexeme L.decimal <* char 'u'

floatLiteral :: Parser Double
floatLiteral = lexeme L.float

reserved :: Text -> Parser ()
reserved word = lexeme (string word *> notFollowedBy alphaNumChar)

reservedWords :: [Text]
reservedWords =
  [ "if",
    "else",
    "while",
    "for",
    "return",
    "int",
    "uint",
    "float",
    "char",
    "bool",
    "struct",
    "sizeof",
    "true",
    "false",
    "fn",
    "on",
    "var"
  ]

identifier :: Parser Text
identifier = (lexeme . try) (p >>= check)
  where
    p = fmap T.pack $ (:) <$> letterChar <*> many (alphaNumChar <|> char '_')
    check x =
      if x `elem` reservedWords
        then fail $ "keyword " <> show x <> " cannot be an identifier"
        else return x

parens :: Parser a -> Parser a
parens = between (symbol "(") (symbol ")")

braces :: Parser a -> Parser a
braces = between (symbol "{") (symbol "}")

brackets :: Parser a -> Parser a
brackets = between (symbol "[") (symbol "]")